/* Copyright 1988 Stephan v. Bechtolsheim */

/* This file is part of the TeXPS Software Package.

The TeXPS Software Package is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY.  No author or distributor
accepts responsibility to anyone for the consequences of using it
or for whether it serves any particular purpose or works at all,
unless he says so in writing.  Refer to the TeXPS Software Package
General Public License for full details.

Everyone is granted permission to copy, modify and redistribute
the TeXPS Software Package, but only under the conditions described in the
TeXPS Software Package General Public License.   A copy of this license is
supposed to have been given to you along with TeXPS Software Package so you
can know your rights and responsibilities.  It should be in a
file named CopyrightLong.  Among other things, the copyright notice
and this notice must be preserved on all copies.  */

/* Some simple RCS related procedures. */

#include <stdio.h>
#include "extfil.h"
#include <sys/dir.h>
#include <sys/file.h>
#if SYS_V == 1
#include <string.h>
#else
#include <strings.h>
#endif

/* Definitions. */
#if SYS_V == 1
#define index  strchr
#define rindex strrchr
#endif

#define TRUE 1
#define FALSE 0

/* Externals. */
extern char *StrcpyAlloc();
extern int FileBusinessVerbose;

/* Forward declarations. */
char *ReturnRcsFileName();

/*
 * RcsClean
 * ********
 * Execute rcsclean on a file.
 * First this program will check whether the file itself exists in
 * the first place or not. If the file does NOT exist, it will
 * return TRUE ("nothing to clean").
 * If the file does exist, an rcsclean with the provided revision number
 * (if it is not empty) will be attempted. If it succeeded, the function will return TRUE.
 * Next an rcsclean with NO version number will be attempted. If it
 * succeeds, the function will return TRUE.
 * Otherwise it will return false.
 *
 * fn: file name of this file.
 * rev: revision number provided.
 * RET: TRUE if the file does not exist, or the rcsclean operation
 *      worked and the file was removed.
 *      FALSE if the file could not be rcsclean-ed.
 */
int
RcsClean(fn, rev)
     char *fn;
     char *rev;
{
  struct stat f_status; /* File status buffer. */
  char system_call[256];
  char *p;

  if (Strlen(fn) == 0)
    Fatal ("RcsClean(): empty file name");

  /* Check whether file exists. */
  if (access(fn, R_OK))
    return (TRUE); /* File does not exist. */

  /* Get status of the file, error if it's not a regular file. */
  if (stat(fn, &f_status))
    Fatal2 ("RcsClean(): cannot stat file \"%s\".", fn);
  if (! f_status.st_mode & S_IFREG)
    Fatal2 ("RcsClean(): file \"%s\" is not a regular file (stat call).", fn);

  /* Now try rcsclean, first with the revision number which was provided,
     if there is such. */
  if (Strlen(rev) != 0) {
    sprintf (system_call, "rcsclean -r%s %s", rev, fn);
    SafeSystemCall (system_call);
    if (access(fn, R_OK))
      return (TRUE); /* File went away. */
  }

  /* Now try a bare rcsclean. */
  sprintf (system_call, "rcsclean %s", fn);
  (void) SafeSystemCall (system_call); /* Ignore status returned. */
  if (access(fn, R_OK))
    return (TRUE); /* File went away. */
  
  /* Bad luck: file still around. */
  return (FALSE);
}

/*
 * FileHasRcsFile
 * **************
 * Check whether file fn has an RCS file associated with it.
 *
 * fn: file name (an ordinary source code file name).
 * RET: TRUE if there is a file fn,v or RCS/fn,v.
 */
int
FileHasRcsFile(fn)
     char *fn;
{
  return (Strlen(ReturnRcsFileName(fn)) != 0);
}

/*
 * ReturnRcsFileName
 * *****************
 * Return the RCS file name of a file, if that file has an RCS file
 * in the first place. If it does NOT have an RCS file, return NULL.
 *
 * fn: file name.
 * RET: RCS file name, NULL if there is no such file.
 */
char *
ReturnRcsFileName(fn)
     char *fn;
{
  char buffer[512];
  char buffer2[512];
  char *p;

  if (Strlen(fn) == 0)
    Fatal ("ReturnRcsFileName(): file name empty.");

  /* The easy case first: check for an RCS file in the
     same directory the file is in. The following return
     is executed, if there is such RCS file. */
  sprintf (buffer, "%s,v", fn);
  if (FileBusinessVerbose)
    fprintf(stderr, "ReturnRcsFileName(): RCS file name[1]: \"%s\"\n", buffer);
  if (access(buffer, R_OK) == 0)
    return (StrcpyAlloc(buffer));

  /* Now we need to look into the RCS subdirectory (if there is
     such: we don't really care: just build the file name and
     use access. */
  strcpy (buffer, fn);

  /* Build in buffer2 the RCS file name which will be used for testing. */
  if ((p=rindex(buffer, '/')) == NULL)
    /* There is NO directory part (no /) in this file name! */
    sprintf(buffer2, "RCS/%s,v", buffer);
  else {
    /* Take the directory part, put a "/RCS/" inbetween and then
       append the remainder of the file name. First split the
       file name into two parts. */
    *p = '\0';
    sprintf (buffer2, "%s/RCS/%s,v", buffer, p+1);
  }

  if (FileBusinessVerbose)
    fprintf(stderr, "ReturnRcsFileName(): RCS file name[2]: \"%s\"\n", buffer2);

  if (access(buffer2, R_OK) == 0)
    return (StrcpyAlloc(buffer2));
  else
    return(NULL);
}
